@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Implementation Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,8 +19,10 @@
 
 #include <string.h>
 #include <gtk/gtk.h>
+#include <gdk/gdkkeysyms.h>
 #include "gailentry.h"
 #include "gailcombo.h"
+#include "gailcombobox.h"
 #include <libgail-util/gailmisc.h>
 
 static void       gail_entry_class_init            (GailEntryClass       *klass);
@@ -127,6 +129,8 @@ static void       gail_entry_paste_received	   (GtkClipboard *clipboard,
 
 /* Callbacks */
 
+static void       gail_entry_notify_insert         (GailEntry            *entry);
+static void       gail_entry_notify_delete         (GailEntry            *entry);
 static void	  _gail_entry_insert_text_cb	   (GtkEntry     	 *entry,
                                                     gchar		 *arg1,
                                                     gint		 arg2,
@@ -146,6 +150,8 @@ static gboolean              idle_do_action              (gpointer        data);
 static gint                  gail_entry_get_n_actions    (AtkAction       *action);
 static G_CONST_RETURN gchar* gail_entry_get_description  (AtkAction       *action,
                                                           gint            i);
+static G_CONST_RETURN gchar* gail_entry_get_keybinding   (AtkAction       *action,
+                                                          gint            i);
 static G_CONST_RETURN gchar* gail_entry_action_get_name  (AtkAction       *action,
                                                           gint            i);
 static gboolean              gail_entry_set_description  (AtkAction       *action,
@@ -240,10 +246,12 @@ static void
 gail_entry_object_init (GailEntry *entry)
 {
   entry->textutil = NULL;
-  entry->signal_name = NULL;
+  entry->signal_name_insert = NULL;
+  entry->signal_name_delete = NULL;
   entry->cursor_position = 0;
   entry->selection_bound = 0;
   entry->activate_description = NULL;
+  entry->activate_keybinding = NULL;
 }
 
 static void
@@ -272,6 +280,11 @@ gail_entry_real_initialize (AtkObject *obj,
 	G_CALLBACK (_gail_entry_delete_text_cb), NULL);
   g_signal_connect (data, "changed",
 	G_CALLBACK (_gail_entry_changed_cb), NULL);
+
+  if (entry->visible)
+    obj->role = ATK_ROLE_TEXT;
+  else
+    obj->role = ATK_ROLE_PASSWORD_TEXT;
 }
 
 static void
@@ -290,14 +303,8 @@ gail_entry_real_notify_gtk (GObject		*obj,
 
   if (strcmp (pspec->name, "cursor-position") == 0)
     {
-      if (entry->signal_name)
-        {
-          g_signal_emit_by_name (atk_obj, 
-                                 entry->signal_name,
-                                 entry->position,
-                                 entry->length);
-          entry->signal_name = NULL;
-        }
+      gail_entry_notify_insert (entry);
+
       if (check_for_selection_change (entry, gtk_entry))
         g_signal_emit_by_name (atk_obj, "text_selection_changed");
       /*
@@ -308,14 +315,8 @@ gail_entry_real_notify_gtk (GObject		*obj,
     }
   else if (strcmp (pspec->name, "selection-bound") == 0)
     {
-      if (entry->signal_name)
-        {
-          g_signal_emit_by_name (atk_obj, 
-                                 entry->signal_name,
-                                 entry->position,
-                                 entry->length);
-          entry->signal_name = NULL;
-        }
+      gail_entry_notify_insert (entry);
+
       if (check_for_selection_change (entry, gtk_entry))
         g_signal_emit_by_name (atk_obj, "text_selection_changed");
     }
@@ -330,14 +331,11 @@ gail_entry_real_notify_gtk (GObject		*obj,
     }
   else if (strcmp (pspec->name, "visibility") == 0)
     {
-      GValue value = { 0, };
       gboolean visibility;
       AtkRole new_role;
 
       text_setup (entry, gtk_entry);
-      g_value_init (&value, G_TYPE_BOOLEAN);
-      g_object_get_property (obj, "visibility", &value);
-      visibility = g_value_get_boolean (&value);
+      visibility = gtk_entry_get_visibility (gtk_entry);
       new_role = visibility ? ATK_ROLE_TEXT : ATK_ROLE_PASSWORD_TEXT;
       atk_object_set_role (atk_obj, new_role);
     }
@@ -388,11 +386,22 @@ gail_entry_finalize (GObject            *object)
 
   g_object_unref (entry->textutil);
   g_free (entry->activate_description);
+  g_free (entry->activate_keybinding);
   if (entry->action_idle_handler)
     {
-      gtk_idle_remove (entry->action_idle_handler);
+      g_source_remove (entry->action_idle_handler);
       entry->action_idle_handler = 0;
     }
+  if (entry->insert_idle_handler)
+    {
+      g_source_remove (entry->insert_idle_handler);
+      entry->insert_idle_handler = 0;
+    }
+  if (entry->delete_idle_handler)
+    {
+      g_source_remove (entry->delete_idle_handler);
+      entry->delete_idle_handler = 0;
+    }
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -409,11 +418,6 @@ gail_entry_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-  if (GTK_ENTRY (widget)->visible)
-    accessible->role = ATK_ROLE_TEXT;
-  else
-    accessible->role = ATK_ROLE_PASSWORD_TEXT;
-
   return accessible;
 }
 
@@ -425,7 +429,8 @@ gail_entry_get_index_in_parent (AtkObject *accessible)
    * otherwise do the normal thing.
    */
   if (accessible->accessible_parent)
-    if (GAIL_IS_COMBO (accessible->accessible_parent))
+    if (GAIL_IS_COMBO (accessible->accessible_parent) ||
+        GAIL_IS_COMBO_BOX (accessible->accessible_parent))
       return 1;
 
   return ATK_OBJECT_CLASS (parent_class)->get_index_in_parent (accessible);
@@ -789,7 +794,7 @@ gail_entry_get_selection (AtkText *text,
   * selection_num is 0.
   */
   if (selection_num != 0)
-     return 0;
+     return NULL;
 
   entry = GTK_ENTRY (widget);
   gtk_editable_get_selection_bounds (GTK_EDITABLE (entry), start_pos, end_pos);
@@ -797,7 +802,7 @@ gail_entry_get_selection (AtkText *text,
   if (*start_pos != *end_pos)
      return gtk_editable_get_chars (GTK_EDITABLE (entry), *start_pos, *end_pos);
   else
-     return 0;
+     return NULL;
 }
 
 static gboolean
@@ -1064,13 +1069,37 @@ gail_entry_paste_received (GtkClipboard *clipboard,
 
 /* Callbacks */
 
+static gboolean
+idle_notify_insert (gpointer data)
+{
+  GailEntry *entry = GAIL_ENTRY (data);
+
+  entry->insert_idle_handler = 0;
+  gail_entry_notify_insert (entry);
+  return FALSE;
+}
+
+static void
+gail_entry_notify_insert (GailEntry *entry)
+{
+  if (entry->signal_name_insert)
+    {
+      g_signal_emit_by_name (entry, 
+                             entry->signal_name_insert,
+                             entry->position_insert,
+                             entry->length_insert);
+      entry->signal_name_insert = NULL;
+    }
+}
+
 /* Note arg1 returns the character at the start of the insert.
  * arg2 returns the number of characters inserted.
  */
-static void _gail_entry_insert_text_cb (GtkEntry *entry, 
-                                        gchar    *arg1, 
-                                        gint     arg2,
-                                        gpointer arg3)
+static void 
+_gail_entry_insert_text_cb (GtkEntry *entry, 
+                            gchar    *arg1, 
+                            gint     arg2,
+                            gpointer arg3)
 {
   AtkObject *accessible;
   GailEntry *gail_entry;
@@ -1078,15 +1107,18 @@ static void _gail_entry_insert_text_cb (GtkEntry *entry,
 
   accessible = gtk_widget_get_accessible (GTK_WIDGET (entry));
   gail_entry = GAIL_ENTRY (accessible);
-  if (!gail_entry->signal_name)
+  if (!gail_entry->signal_name_insert)
     {
-      gail_entry->signal_name = "text_changed::insert";
-      gail_entry->position = *position;
-      gail_entry->length = arg2;
+      gail_entry->signal_name_insert = "text_changed::insert";
+      gail_entry->position_insert = *position;
+      gail_entry->length_insert = arg2;
     }
   /*
    * The signal will be emitted when the cursor position is updated.
+   * or in an idle handler if it not updated.
    */
+   if (gail_entry->insert_idle_handler == 0)
+     gail_entry->insert_idle_handler = g_idle_add (idle_notify_insert, gail_entry);
 }
 
 static gunichar 
@@ -1121,16 +1153,39 @@ gail_entry_get_character_at_offset (AtkText *text,
   return unichar;
 }
 
+static gboolean
+idle_notify_delete (gpointer data)
+{
+  GailEntry *entry = GAIL_ENTRY (data);
+
+  entry->delete_idle_handler = 0;
+  gail_entry_notify_delete (entry);
+  return FALSE;
+}
+
+static void
+gail_entry_notify_delete (GailEntry *entry)
+{
+  if (entry->signal_name_delete)
+    {
+      g_signal_emit_by_name (entry, 
+                             entry->signal_name_delete,
+                             entry->position_delete,
+                             entry->length_delete);
+      entry->signal_name_delete = NULL;
+    }
+}
+
 /* Note arg1 returns the start of the delete range, arg2 returns the
  * end of the delete range if multiple characters are deleted.	
  */
-static void _gail_entry_delete_text_cb (GtkEntry *entry, 
-                                        gint      arg1, 
-                                        gint      arg2)
+static void 
+_gail_entry_delete_text_cb (GtkEntry *entry, 
+                            gint      arg1, 
+                            gint      arg2)
 {
   AtkObject *accessible;
-
-  accessible = gtk_widget_get_accessible (GTK_WIDGET (entry));
+  GailEntry *gail_entry;
 
   /*
    * Zero length text deleted so ignore
@@ -1138,7 +1193,20 @@ static void _gail_entry_delete_text_cb (GtkEntry *entry,
   if (arg2 - arg1 == 0)
     return;
 
-  g_signal_emit_by_name (accessible, "text_changed::delete", arg1, arg2 - arg1);
+  accessible = gtk_widget_get_accessible (GTK_WIDGET (entry));
+  gail_entry = GAIL_ENTRY (accessible);
+  if (!gail_entry->signal_name_delete)
+    {
+      gail_entry->signal_name_delete = "text_changed::delete";
+      gail_entry->position_delete = arg1;
+      gail_entry->length_delete = arg2 - arg1;
+    }
+  /*
+   * The signal will be emitted when the cursor position is updated.
+   * or in an idle handler if it not updated.
+   */
+   if (gail_entry->delete_idle_handler == 0)
+     gail_entry->delete_idle_handler = g_idle_add (idle_notify_delete, gail_entry);
 }
 
 static void
@@ -1192,6 +1260,7 @@ atk_action_interface_init (AtkActionIface *iface)
   iface->do_action = gail_entry_do_action;
   iface->get_n_actions = gail_entry_get_n_actions;
   iface->get_description = gail_entry_get_description;
+  iface->get_keybinding = gail_entry_get_keybinding;
   iface->get_name = gail_entry_action_get_name;
   iface->set_description = gail_entry_set_description;
 }
@@ -1221,7 +1290,7 @@ gail_entry_do_action (AtkAction *action,
       if (entry->action_idle_handler)
         return_value = FALSE;
       else
-        entry->action_idle_handler = gtk_idle_add (idle_do_action, entry);
+        entry->action_idle_handler = g_idle_add (idle_do_action, entry);
       break;
     default:
       return_value = FALSE;
@@ -1278,6 +1347,73 @@ gail_entry_get_description (AtkAction *action,
   return return_value; 
 }
 
+static G_CONST_RETURN gchar*
+gail_entry_get_keybinding (AtkAction *action,
+                           gint      i)
+{
+  GailEntry *entry;
+  gchar *return_value = NULL;
+
+  entry = GAIL_ENTRY (action);
+  switch (i)
+    {
+    case 0:
+      {
+        /*
+         * We look for a mnemonic on the label
+         */
+        GtkWidget *widget;
+        GtkWidget *label;
+        AtkRelationSet *set;
+        AtkRelation *relation;
+        GPtrArray *target;
+        gpointer target_object;
+        guint key_val; 
+
+        entry = GAIL_ENTRY (action);
+        widget = GTK_ACCESSIBLE (entry)->widget;
+        if (widget == NULL)
+          /*
+           * State is defunct
+           */
+          return NULL;
+
+        /* Find labelled-by relation */
+
+        set = atk_object_ref_relation_set (ATK_OBJECT (action));
+        if (!set)
+          return NULL;
+        label = NULL;
+        relation = atk_relation_set_get_relation_by_type (set, ATK_RELATION_LABELLED_BY);
+        if (relation)
+          {              
+            target = atk_relation_get_target (relation);
+          
+            target_object = g_ptr_array_index (target, 0);
+            if (GTK_IS_ACCESSIBLE (target_object))
+              {
+                label = GTK_ACCESSIBLE (target_object)->widget;
+              } 
+          }
+
+        g_object_unref (set);
+
+        if (GTK_IS_LABEL (label))
+          {
+            key_val = gtk_label_get_mnemonic_keyval (GTK_LABEL (label)); 
+            if (key_val != GDK_VoidSymbol)
+              return_value = gtk_accelerator_name (key_val, GDK_MOD1_MASK);
+          }
+        g_free (entry->activate_keybinding);
+        entry->activate_keybinding = return_value;
+        break;
+      }
+    default:
+      break;
+    }
+  return return_value; 
+}
+
 static G_CONST_RETURN gchar*
 gail_entry_action_get_name (AtkAction *action,
                             gint      i)
