@@ -29,7 +29,7 @@ static AtkObject*            gail_item_ref_child       (AtkObject     *obj,
                                                         gint          i);
 static void                  gail_item_real_initialize (AtkObject     *obj,
                                                         gpointer      data);
-static gint                  gail_item_label_map_gtk   (GtkWidget     *widget,
+static void                  gail_item_label_map_gtk   (GtkWidget     *widget,
                                                         gpointer      data);
 static void                  gail_item_finalize        (GObject       *object);
 static void                  gail_item_init_textutil   (GailItem      *item,
@@ -80,7 +80,7 @@ static AtkAttributeSet* gail_item_get_run_attributes
 					            gint	      *end_offset);
 static AtkAttributeSet* gail_item_get_default_attributes
                                                    (AtkText           *text);
-static GtkWidget*            get_label_from_item   (GtkWidget         *item);
+static GtkWidget*            get_label_from_container   (GtkWidget    *container);
 
 
 
@@ -174,21 +174,24 @@ gail_item_real_initialize (AtkObject *obj,
   ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
   item->textutil = NULL;
+  item->text = NULL;
 
-  label = get_label_from_item (GTK_WIDGET (data));
+  label = get_label_from_container (GTK_WIDGET (data));
   if (GTK_IS_LABEL (label))
-    if (GTK_WIDGET_MAPPED (label))
-      gail_item_init_textutil (item, label);
-    else
-      g_signal_connect (label,
-                        "map",
-                        G_CALLBACK (gail_item_label_map_gtk),
-                        item);
+    {
+      if (GTK_WIDGET_MAPPED (label))
+        gail_item_init_textutil (item, label);
+      else
+        g_signal_connect (label,
+                          "map",
+                          G_CALLBACK (gail_item_label_map_gtk),
+                          item);
+    }
 
   obj->role = ATK_ROLE_LIST_ITEM;
 }
 
-static gint
+static void
 gail_item_label_map_gtk (GtkWidget *widget,
                          gpointer data)
 {
@@ -225,6 +228,10 @@ gail_item_finalize (GObject *object)
     {
       g_object_unref (item->textutil);
     }
+  if (item->text)
+    {
+      g_free (item->text);
+    }
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -236,9 +243,7 @@ gail_item_get_name (AtkObject *obj)
   g_return_val_if_fail (GAIL_IS_ITEM (obj), NULL);
 
   name = ATK_OBJECT_CLASS (parent_class)->get_name (obj);
-  if (name != NULL)
-    return name;
-  else
+  if (name == NULL)
     {
       /*
        * Get the label child
@@ -253,7 +258,7 @@ gail_item_get_name (AtkObject *obj)
          */
         return NULL;
 
-      label = get_label_from_item (widget);
+      label = get_label_from_container (widget);
       if (GTK_IS_LABEL (label))
 	return gtk_label_get_text (GTK_LABEL(label));
       /*
@@ -261,26 +266,63 @@ gail_item_get_name (AtkObject *obj)
        * the label of the selected item is detached from the menu item
        */
       else if (GTK_IS_MENU_ITEM (widget))
-         {
-           GtkWidget *parent;
-           GtkWidget *attach;
+        {
+          GtkWidget *parent;
+          GtkWidget *attach;
+          GList *list;
+          AtkObject *parent_obj;
+          gint index;
 
-           parent = gtk_widget_get_parent (widget);
-           if (GTK_IS_MENU (parent))
-             {
-               attach = gtk_menu_get_attach_widget (GTK_MENU (parent)); 
+          parent = gtk_widget_get_parent (widget);
+          if (GTK_IS_MENU (parent))
+            {
+              attach = gtk_menu_get_attach_widget (GTK_MENU (parent)); 
 
-               if (GTK_IS_OPTION_MENU (attach))
-                 {
-                   label = get_label_from_item (attach);
-                   if (GTK_IS_LABEL (label))
-	             return gtk_label_get_text (GTK_LABEL(label));
-                 }
-             }
-         }
+              if (GTK_IS_OPTION_MENU (attach))
+                {
+                  label = get_label_from_container (attach);
+                  if (GTK_IS_LABEL (label))
+	            return gtk_label_get_text (GTK_LABEL(label));
+                }
+              list = gtk_container_get_children (GTK_CONTAINER (parent));
+              index = g_list_index (list, widget);
+              g_list_free (list);
+              parent_obj = atk_object_get_parent (gtk_widget_get_accessible (parent));
+              if (GTK_IS_ACCESSIBLE (parent_obj))
+                {
+                  parent = GTK_ACCESSIBLE (parent_obj)->widget;
+                  if (GTK_IS_COMBO_BOX (parent))
+                    {
+                      GtkTreeModel *model;
+                      GtkTreeIter iter;
+                      GailItem *item;
+                      gint n_columns, i;
 
-      return NULL;
+                      model = gtk_combo_box_get_model (GTK_COMBO_BOX (parent));                       
+                      item = GAIL_ITEM (obj);
+                      g_free (item->text);
+                      if (gtk_tree_model_iter_nth_child (model, &iter, NULL, index))
+                        {
+                          n_columns = gtk_tree_model_get_n_columns (model);
+                          for (i = 0; i < n_columns; i++)
+                            {
+                              GValue value = { 0, };
+
+                               gtk_tree_model_get_value (model, &iter, i, &value);
+                               if (G_VALUE_HOLDS_STRING (&value))
+                                 {
+                                   item->text =  (gchar *) g_value_get_string (&value);
+                                   break;
+                                 }
+                            }
+                        }
+                      name = item->text;
+                    }
+                }
+            }
+        }
     }
+  return name;
 }
 
 /*
@@ -367,7 +409,7 @@ gail_item_get_text (AtkText *text,
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL (label))
     return NULL;
@@ -405,7 +447,7 @@ gail_item_get_text_before_offset (AtkText         *text,
     return NULL;
   
   /* Get label */
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -437,7 +479,7 @@ gail_item_get_text_at_offset (AtkText         *text,
     return NULL;
   
   /* Get label */
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -471,7 +513,7 @@ gail_item_get_text_after_offset (AtkText         *text,
   }
   
   /* Get label */
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -496,7 +538,7 @@ gail_item_get_character_count (AtkText *text)
     /* State is defunct */
     return 0;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return 0;
@@ -525,7 +567,7 @@ gail_item_get_character_extents (AtkText      *text,
     /* State is defunct */
     return;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return;
@@ -555,7 +597,7 @@ gail_item_get_offset_at_point (AtkText      *text,
     /* State is defunct */
     return -1;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return -1;
@@ -594,7 +636,7 @@ gail_item_get_run_attributes (AtkText *text,
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -636,7 +678,7 @@ gail_item_get_default_attributes (AtkText *text)
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -661,7 +703,7 @@ gail_item_get_character_at_offset (AtkText *text,
     /* State is defunct */
     return '\0';
 
-  label = get_label_from_item (widget);
+  label = get_label_from_container (widget);
 
   if (!GTK_IS_LABEL(label))
     return '\0';
@@ -674,15 +716,15 @@ gail_item_get_character_at_offset (AtkText *text,
 }
 
 static GtkWidget*
-get_label_from_item (GtkWidget *item)
+get_label_from_container (GtkWidget *container)
 {
   GtkWidget *label;
   GList *children, *tmp_list;
 
-  if (!GTK_IS_CONTAINER (item))
+  if (!GTK_IS_CONTAINER (container))
     return NULL;
  
-  children = gtk_container_get_children (GTK_CONTAINER (item));
+  children = gtk_container_get_children (GTK_CONTAINER (container));
   label = NULL;
 
   for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
@@ -692,6 +734,16 @@ get_label_from_item (GtkWidget *item)
            label = tmp_list->data;
            break;
 	}
+      /*
+       * Get label from menu item in desktop background preferences
+       * option menu. See bug #144084.
+       */
+      else if (GTK_IS_BOX (tmp_list->data))
+        {
+           label = get_label_from_container (GTK_WIDGET (tmp_list->data));
+           if (label)
+             break;
+        }
     }
   g_list_free (children);
 
